# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Foobar.
# If not, see <https://www.gnu.org/licenses/>.

import os
import time
import tempfile
import queue
import threading
import multiprocessing

import pytest

import octopus_sensing.devices.audio_streaming as audio_streaming
from octopus_sensing.common.message_creators import start_message, stop_message, terminate_message


class MockedMiniAudioModule:

    def Devices(self):
        return MockedMiniAudioDevices()

    def CaptureDevice(self, buffersize_msec, sample_rate, device_id):
        return MockedMiniAudioCapture()

    def DecodedSoundFile(self, name, nchannels, sample_rate, file_format, samples):
        # Just returning back samples without change!
        return samples

    def wav_write_file(self, file_name, sound_data):
        # Just writing our fake data to the file.
        open(file_name, 'w').write(str(sound_data))


class MockedMiniAudioDevices:

    def get_captures(self):
        # A dict of device_id to a Device Object.
        # From the Device Object, we only use 'id' field.
        return {1: {'id': 'dev-1'}}


class MockedMiniAudioCapture:

    # None of these are important. They will just pass back to our Mock.
    nchannels = 5
    sample_rate = 125
    format = 'wav'

    def start(self, recorder):
        recorder.send(b'123')

    def stop(self):
        pass


@pytest.fixture(scope="module")
def mocked():
    # Replacing base with Thread instead of Process. Because:
    # 1. Process on Windows doesn't use fork (copy-on-write) so we will lose
    #    the mocks we made in here (the parent process).
    # 2. Coverage will lose track in the child process.
    original_bases = audio_streaming.AudioStreaming.__bases__[0].__bases__
    audio_streaming.AudioStreaming.__bases__[0].__bases__ = (threading.Thread,)

    # We're replacing the 'miniaudio' imported by our 'audio_streaming' module
    # with our mocked one.
    original_miniaudio = audio_streaming.miniaudio
    audio_streaming.miniaudio = MockedMiniAudioModule()

    yield None

    # Replacing back the original things
    audio_streaming.AudioStreaming.__bases__[0].__bases__ = original_bases
    audio_streaming.miniaudio = original_miniaudio


def test_happy_path(mocked):

    output_dir = tempfile.mkdtemp(prefix="octopus-sensing-test")
    device_name = 'test-audio-device'
    experiment_id = 'test-exp-1'
    stimuli_id = 'sti-1'

    device = audio_streaming.AudioStreaming(
        1, name=device_name, output_path=output_dir)

    # Since there's no device coordinator running, we set the queue ourselves,
    # and will start the process.
    msg_queue = queue.Queue()
    device.set_queue(msg_queue)
    device.start()
    # To ensure the process is started.
    time.sleep(0.2)

    msg_queue.put(start_message(experiment_id, stimuli_id))

    time.sleep(1)

    msg_queue.put(stop_message(experiment_id, stimuli_id))

    time.sleep(0.2)

    # It should save the file after receiving a STOP.
    device_output = os.path.join(output_dir, device_name)
    assert os.path.exists(device_output)
    recorded_file = os.path.join(device_output, '{}-{}-{}.wav'.format(
        device_name, experiment_id, stimuli_id))
    assert os.path.exists(recorded_file)

    # Sending terminate and waiting for the device process to exit.
    msg_queue.put(terminate_message())
    device.join()



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

import threading
import socket
from typing import List, Optional

from octopus_sensing.common.message_creators import MessageType
from octopus_sensing.common.message import Message
from octopus_sensing.devices.device import Device


class SocketNetworkDevice(Device):
    '''
    This class is being used for sending triggers to other softwares using TCP-IP socket.
    It works as a server socket, which sends triggers (when an event happen) to the connected clients.

    For example if we have a device that record data through matlab, through this server socket, we can
    send the trigger to the matlab application to mark the recorded data.

    Attributes
    ----------

    Parameters
    ----------
    host: str
        host IP address

    port: str
        port number

    Example
    -------
    Creating a SocketNetworkDevice in the local machine and adding it to the device_coordinator.
    By adding it to the DeviceCoordinator, it starts listening

    >>> device_coordinator = DeviceCoordinator()
    >>> socket_device = SocketNetworkDevice("0.0.0.0", 5002)
    >>> device_coordinator.add_devices([socket_device])

    Note
    -----
    Look at Examples/send_trigger_to_remote_device.py (server code),
    Examples/matlabRecorder.m or examples/client.py (client codes)

    '''

    def __init__(self,
                 host: str,
                 port: str,
                 **kwargs):
        super().__init__(**kwargs)

        self._host = host
        self._port = port
        self._server_socket: Optional[socket.socket] = None
        self.__connections: List[socket.socket] = []
        self._stop_listening = False
        self._trigger: Optional[str] = None
        self._experiment_id: Optional[str] = None

    def _accept_connections(self):
        self._server_socket.listen(5)
        while True:
            if self._stop_listening is True:
                break
            try:
                conn, address = self._server_socket.accept()  # accept new connection
            except socket.timeout:
                continue
            if conn is not None:
                self.__connections.append(conn)

        self._server_socket.shutdown(socket.SHUT_RDWR)
        self._server_socket.close()

    def _run(self):
        self._server_socket = socket.socket()  # get instance
        self._server_socket.setsockopt(
            socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._server_socket.settimeout(0.5)
        # bind host address and port together
        self._server_socket.bind((self._host, self._port))

        threading.Thread(target=self._accept_connections).start()
        while True:
            message = self.message_queue.get()
            if message is None:
                continue
            if message.type == MessageType.START:
                self._experiment_id = message.experiment_id
                self.__set_trigger(message)
                for connection in self.__connections:
                    threading.Thread(target=self.__send_message,
                                     args=(connection,)).start()

            elif message.type == MessageType.STOP:
                self._experiment_id = message.experiment_id
                self.__set_trigger(message)
                for connection in self.__connections:
                    threading.Thread(target=self.__send_message,
                                     args=(connection,)).start()

            elif message.type == MessageType.TERMINATE:
                self._trigger = "terminate"
                self._stop_listening = True
                for connection in self.__connections:
                    threading.Thread(target=self.__send_message,
                                     args=(connection,)).start()
                break

    def __send_message(self, connection: socket.socket):
        '''
        Gets a connection and sends the trigger to it

        Parameters
        ----------
        connection: socket.socket
            a socket connection
        '''
        assert self._trigger is not None
        self._trigger += "\n"
        connection.send(self._trigger.encode())

    def __set_trigger(self, message: Message):
        '''
        Takes a message and set the trigger using its data

        Parameters
        ----------
        message: Message
            a message object
        '''
        self._trigger = \
            "{0}-{1}-{2}".format(message.type,
                                 message.experiment_id,
                                 str(message.stimulus_id).zfill(2))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['octopus_sensing',
 'octopus_sensing.common',
 'octopus_sensing.devices',
 'octopus_sensing.devices.network_devices',
 'octopus_sensing.preprocessing',
 'octopus_sensing.questionnaire',
 'octopus_sensing.stimuli',
 'octopus_sensing.tests',
 'octopus_sensing.windows']

package_data = \
{'': ['*'],
 'octopus_sensing': ['OpenVibe/*'],
 'octopus_sensing.tests': ['data/preprocess_expected/OpenBCI_16_continuous/*',
                           'data/preprocess_expected/OpenBCI_16_sep/*',
                           'data/preprocess_expected/OpenBCI_8_continuous/*',
                           'data/preprocess_expected/OpenBCI_8_sep/*',
                           'data/preprocess_expected/Shimmer_continuous/gsr/*',
                           'data/preprocess_expected/Shimmer_continuous/ppg/*',
                           'data/preprocess_expected/Shimmer_sep/gsr/*',
                           'data/preprocess_expected/Shimmer_sep/ppg/*',
                           'data/recorded/OpenBCI_16_continuous/*',
                           'data/recorded/OpenBCI_16_sep/*',
                           'data/recorded/OpenBCI_8_continuous/*',
                           'data/recorded/OpenBCI_8_sep/*',
                           'data/recorded/Shimmer_continuous/*',
                           'data/recorded/Shimmer_sep/*']}

install_requires = \
['bitstring>=3.1.7,<4.0.0',
 'brainflow>=4.6.1,<5.0.0',
 'heartpy>=1.2.7,<2.0.0',
 'miniaudio>=1.45,<2.0',
 'mne>=0.23.0,<0.24.0',
 'msgpack>=1.0.0,<2.0.0',
 'opencv-python>=4.4.0,<5.0.0',
 'pandas>=1.2.4,<2.0.0',
 'pyOpenBCI>=0.13,<0.14',
 'pyserial>=3.4,<4.0',
 'requests>=2.24.0,<3.0.0',
 'scipy>=1.5.2,<2.0.0',
 'screeninfo>=0.7,<0.8',
 'sounddevice>=0.4.0,<0.5.0',
 'xmltodict>=0.12.0,<0.13.0']

extras_require = \
{':sys_platform != "win32"': ['PyGObject==3.40.1'],
 ':sys_platform == "linux"': ['bluepy>=1.3.0,<2.0.0']}

setup_kwargs = {
    'name': 'octopus-sensing',
    'version': '3.1.1',
    'description': 'Library for recording data synchronously from different physiological sensors',
    'long_description': '<h2 align="center">Octopus Sensing</h2>\n<p align="center">\n  <img src="https://octopus-sensing.nastaran-saffar.me/_static/octopus-sensing-logo-small.png" alt="Octopus Sensing Logo">\n</p>\n<p align="center">\n  <img src="https://img.shields.io/github/workflow/status/octopus-sensing/octopus-sensing/Python%20Check?label=checks" alt="GitHub Workflow Status">\n  <img src="https://img.shields.io/codecov/c/gh/octopus-sensing/octopus-sensing" alt="Codecov">\n  <img src="https://img.shields.io/pypi/v/octopus-sensing" alt="PyPI">\n  <img src="https://img.shields.io/pypi/l/octopus-sensing" alt="PyPI - License">\n</p>\n\nOctopus Sensing is a tool to help you run scientific experiments that involve recording data synchronously from\nmultiple sources in human-computer interaction studies. You write steps of an experiment scenario, for example showing a stimulus and then a questionnaire. The tool takes care of the rest.\n\nIt can collect data from multiple devices such as OpenBCI EEG headset, Shimmer sensor (GSR and PPG),\nVideo and Audio and so forth simultaneously. Data collection can be started and stopped synchronously across all devices.\nCollected data will be tagged with the timestamp of the start and stop of the experiment, the ID of\nthe experiment, etc.\n\nThe aim is to make the scripting interface so simple that people with minimum or no software\ndevelopment skills can define experiment scenarios with no effort.\nAlso, this tool can be used as the base structure for creating real-time data processing systems like systems with capabilities of recognizing emotions, stress, cognitive load, or analyzing human behaviors.\n\n\n**To see the full documentation visit the [Octopus Sensing website](https://octopus-sensing.nastaran-saffar.me/).**\n\nMain features\n--------------\n\n* Controls data recording from multiple sources using a simple unified interface\n* Tags an event on collected data, such as the start of an experiment, and events during the experiment, etc.\n* Can show stimuli (images and videos) and questionnaires\n* Monitoring interface that visualizes collected data in real-time\n* Offline visualization of data from multiple sources simultanously\n\nCopyright\n---------\nCopyright © 2020-2022 Nastaran Saffaryazdi, Aidin Gharibnavaz\n\nThis program is free software: you can redistribute it and/or modify it under the terms of the GNU\nGeneral Public License as published by the Free Software Foundation, either version 3 of the\nLicense, or (at your option) any later version.\n\nSee [License file](https://github.com/nastaran62/octopus-sensing/blob/master/LICENSE) for full terms.\n',
    'author': 'Nastaran Saffaryazdi',
    'author_email': 'nsaffar@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://octopus-sensing.nastaran-saffar.me',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<4.0',
}


setup(**setup_kwargs)



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Foobar.
# If not, see <https://www.gnu.org/licenses/>.

import time
import threading
import http.server
import json

import pytest

from octopus_sensing.device_coordinator import DeviceCoordinator
from octopus_sensing.devices.network_devices.http_device import HttpNetworkDevice
from octopus_sensing.common.message_creators import start_message, stop_message


class Handler(http.server.BaseHTTPRequestHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._expected_messages = [
            {"type": "START", "experiment_id": "exp1", "stimulus_id": "stim1"},
            {"type": "STOP", "experiment_id": "exp1", "stimulus_id": "stim2"},
            {"type": "TERMINATE", "experiment_id": "", "stimulus_id": ""},
        ]

    def do_POST(self):
        content_length = int(self.headers.get("Content-Length", "-1"))
        serialized_body = self.rfile.read(content_length)
        body = json.loads(serialized_body)

        assert body == self._expected_messages.pop()

        self.send_response(200)
        self.end_headers()


def test_http_network_device_happy_path():
    coordinator = DeviceCoordinator()
    device = HttpNetworkDevice(
        ["http://localhost:5003/"], name="test-http-network-device", timeout=15)
    coordinator.add_device(device)

    server = http.server.ThreadingHTTPServer(("localhost", 5003), Handler)
    threading.Thread(target=server.serve_forever, daemon=True).start()

    # To ensure device and server are started
    time.sleep(1)

    coordinator.dispatch(start_message("exp1", "stim1"))
    coordinator.dispatch(stop_message("exp1", "stim2"))
    # Should send TERMINATE message
    coordinator.terminate()


def test_validations():
    # Invalid URLs
    with pytest.raises(RuntimeError):
        HttpNetworkDevice(["localhost:5003"])
    with pytest.raises(RuntimeError):
        HttpNetworkDevice(["127.0.0.1"])
    with pytest.raises(RuntimeError):
        HttpNetworkDevice(["ftp://localhost:5003"])
    with pytest.raises(RuntimeError):
        HttpNetworkDevice(["http:/192.168.1.1/"])
    with pytest.raises(RuntimeError):
        HttpNetworkDevice(["http//192.168.1.1/"])
    # Can't really detect port is not a number. This port will be simply ignored!
    # with pytest.raises(RuntimeError):
    #    HttpNetworkDevice(["http://192.168.1.1:abc/"])

    # Valid URLs
    HttpNetworkDevice(
        ["http://localhost:80",
         "https://localhost/",
         "http://192.168.1.1/",
         "http://192.168.1.1:8080/",
         "http://[::1]/",
         "http://[2345:425:2ca1:0000:0000:567:5673:23b5]/",
         "http://[2345:425:2ca1:0000:0000:567:5673:23b5]:8080/",
         "https://localhost/abc/def",
         "http://localhost:9090/data?q=123",
         ])



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

import sys
import pickle
import multiprocessing.queues
import threading
import traceback
from typing import List, Any

from octopus_sensing.devices.device import Device

QueueType = multiprocessing.queues.Queue


class MonitoredDevice(Device):
    '''
    Provides functionalities for monitoring a device's data. 
    For example, visualizing data in real time.
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._monitor_in_q = None
        self._monitor_out_q = None

    def set_monitoring_queues(self, monitor_in_q: QueueType, monitor_out_q: QueueType) -> None:
        '''Sets the queues for communicating with the parent process.
        It should be called before the start of the process.
        '''
        self._monitor_in_q = monitor_in_q
        self._monitor_out_q = monitor_out_q

    def run(self) -> None:
        # Ensuring queues are set.
        assert self._monitor_in_q is not None
        assert self._monitor_out_q is not None

        threading.Thread(target=self._monitor_loop,
                         name=self.__class__.__name__ + " monitor thread", daemon=True) \
            .start()

        self._run()

    def _monitor_loop(self) -> None:
        while True:
            # We don't use the data from the queue. It's just a signal.
            self._monitor_in_q.get()

            try:
                # Only 10ms timeout, because we don't want to take cpu time from the
                # main thread (data collector)
                self._monitor_out_q.put(
                    pickle.dumps(
                        self._get_monitoring_data(),
                        protocol=pickle.HIGHEST_PROTOCOL),
                    timeout=0.01)

            except pickle.PickleError:
                print("Error pickling monitoring data", file=sys.stderr)
                traceback.print_exc()
                # We don't want to keep the parent process waiting
                self._monitor_out_q(pickle.dumps(
                    [], protocol=pickle.HIGHEST_PROTOCOL))

    def _get_monitoring_data(self) -> List[Any]:
        '''Subclasses must implmenet this method. It should return
        a list of latest collected records.
        This method will be called in a separate thread, and should
        be thread-safe.
        Also, implmentation must return as quick as possible, to prevent
        blocking of the main thread that doing the collecting.

        @param requested_records: Number of records that should be returned.
        @type requested_records: int

        @return: List of records, or empty list if there's nothing.
        @rtype: List[Any]
        '''
        raise NotImplementedError()



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

import datetime
import csv
import numpy as np
from typing import List, Any, Tuple


def load_all_samples(file_path: str, channels_cols: Tuple[int, int], time_stamp_col: int, time_format: str):
    '''
    Reads the recorded data file and separate data according to the START and STOP triggers

    Parameters
    ----------
    file_path: str
        The path of recorded data
    
    channels_cols: Tuple[int, int]
        The start column and end column number of channels. 
        For example [1, 16] means column 1 to 16 in the csv file includes channels data
    
    time_stamp_col: int
        The column number of time stamp
    
    time_format: str
        The format of recorded times

    Returns
    ---------
    trial_data, converted_times: tuple(List[Any], List[datetime.datetime])

    trial_data: List[Any]
        A list of a trial's data
    
    converted_times: List[datetime.datetime]
        A list of trial's time stamps

    '''
    data = []
    times = []
    with open(file_path, 'r') as file:
        reader = csv.reader(file, delimiter=',')
        i = 0
        for row in reader:
            if i == 0:
                # Skip header line
                i += 1
                continue
            data.append(np.array(row[channels_cols[0]:channels_cols[1]], dtype=np.float32))
            times.append(row[time_stamp_col])
    converted_times = str_to_times(times, time_format)
    return data, converted_times


def load_all_trials(file_path: str, channels_cols: Tuple[int, int], time_stamp_col: int, triger_col: int, time_format: str):
    '''
    Reads the recorded data files and separate data according to the START and STOP triggers

    Parameters
    ----------
    file_path: str
        The path of recorded data
    
    channels_cols: Tuple[int]
        The start column and end column number of channels. 
        For example [1, 16] means column 1 to 16 in the csv file includes channels data
    
    time_stamp_col: int
        The column number of time stamp
    
    triger_col: int
        The column number of trigger
    
    time_format: str
        The format of recorded times

    Returns
    ---------
    all_trials_data, all_trials_times, trial_numbers: 
        tuple(List[Any], List[List[datetime.datetime]], List[List[int]])

    all_trials_data: List[Any]
        A list of all trials data
    
    all_trials_times: List[List[datetime.datetime]]
        A list of all trials time stamps
    
    all_trials_times: List[List[int]]
        A list of all trials IDs
    '''
    all_trials_data = []
    all_trials_times = []
    trial_numbers = []

    action_flag = None
    data:List[Any] = []
    times: List[Any] = []
    with open(file_path, 'r') as file:
        reader = csv.reader(file, delimiter=',')

        for i, row in enumerate(reader):
            if i == 0:
                # Ignore header
                continue
            if len(row) > triger_col:
                if row[triger_col] not in (None, ''):
                    triger = row[triger_col]
                    action = triger[0:4]
                    trial_number = int(triger[-2:])
                    if action == "STAR":
                        action_flag = 1
                    elif action == "STOP":
                        action_flag = 0
                        all_trials_data.append(data)
                        converted_times = str_to_times(times, time_format)
                        all_trials_times.append(converted_times)
                        trial_numbers.append(trial_number)
                        data = []
                        times = []
            if action_flag == 1:
                data.append(np.array(row[channels_cols[0]:channels_cols[1]], dtype=np.float32))
                times.append(row[time_stamp_col])
    return all_trials_data, all_trials_times, trial_numbers


def str_to_times(times: List[str], time_format: str):
    '''
    Convert a list of str times to datetime

    Parameters
    ----------
    times: List[str]
        A list of times in str format
    
    time_format: str:
        Fomatting time style
    
    Returns
    --------
    List[datetime.datetime]
        A list of times
    '''
    converted_times = []
    for item in times:
        try:
            time = datetime.datetime.strptime(item, time_format)
        except:
            try:
                time = datetime.datetime.strptime(item, '%H:%M:%S')
            except:
                time = datetime.datetime.strptime(item, '%Y-%m-%d %H:%M:%S')
        converted_times.append(time)
    return converted_times


def resample(data: List[Any], times: List[datetime.datetime], sampling_rate: int):
    '''
    Resamples data according to time stamps and sampling rate

    Parameters
    ----------
    data: List[Any]
        data to be resampled
    
    times: List[datetime.datetime]
        A list of timestamp. There is a timestamp for each sample of data

    sampling_rate: int
        Data will be resampled to this sampling rate
    
    Returns
    -------
    numpy.array
        An array of resampled data
    '''
    i = 0
    time_delta = datetime.timedelta(0, 1, 0)
    start_time = times[0]
    sample_start_time = start_time
    block: List[Any] = []
    all_data = []
    j = 0
    block_number = 0
    for item in data:
        if times[i] - sample_start_time > time_delta:
            block_number += 1
            if len(block) < sampling_rate:
                while len(block) < sampling_rate:
                    # repeating last item
                    block.extend(block[-(sampling_rate-len(block)):])
            elif len(block) > sampling_rate:
                # Ignoring last items
                block = block[:sampling_rate]
            all_data.extend(block)
            block = []
            sample_start_time = start_time + (block_number * time_delta)
        block.append(item)
        i += 1
    if len(block) > 0:
        if len(block) > sampling_rate/2:
            while len(block) < sampling_rate:
                # repeating last item
                block.extend(block[-(sampling_rate-len(block)):])
            block_number += 1
            all_data.extend(block)
    return np.array(all_data)


def load_all_samples_without_time(file_path: str, channels_cols: Tuple[int, int]):
    '''
    Reads the recorded data file and separate data according to the START and STOP triggers

    Parameters
    ----------
    file_path: str
        The path of recorded data
    
    channels_cols: Tuple[int, int]
        The start column and end column number of channels. 
        For example [1, 16] means column 1 to 16 in the csv file includes channels data


    Returns
    ---------
    trial_data: List[Any]
        A list of a trial's data
    '''
    data = []
    with open(file_path, 'r') as file:
        reader = csv.reader(file, delimiter=',')
        for row in reader:
            data.append(np.array(row[channels_cols[0]:channels_cols[1]], dtype=np.float32))
    return data


def load_all_trials_without_time(file_path: str, channels_cols: Tuple[int, int], triger_col: int):
    '''
    Reads the recorded data files and separate data according to the START and STOP triggers

    Parameters
    ----------
    file_path: str
        The path of recorded data
    
    channels_cols: Tuple[int, int]
        The start column and end column number of channels. 
        For example [1, 16] means column 1 to 16 in the csv file includes channels data
    
    triger_col: int
        The column number of trigger

    Returns
    ---------
    all_trials_data, trial_numbers: 
        tuple(List[Any], List[List[int]])

    all_trials_data: List[Any]
        A list of all trials data
    
    all_trials_times: List[List[int]]
        A list of all trials IDs
    '''
    all_trials_data = []
    trial_numbers = []

    action_flag = None
    data:List[Any] = []
    with open(file_path, 'r') as file:
        reader = csv.reader(file, delimiter=',')

        for i, row in enumerate(reader):
            if len(row) > triger_col:
                if row[triger_col] not in (None, ''):
                    triger = row[triger_col]
                    action = triger[0:4]
                    trial_number = int(triger[-2:])
                    if action == "STAR":
                        action_flag = 1
                    elif action == "STOP":
                        action_flag = 0
                        all_trials_data.append(data)
                        trial_numbers.append(trial_number)
                        data = []
            if action_flag == 1:
                data.append(np.array(row[channels_cols[0]:channels_cols[1]], dtype=np.float32))
    return all_trials_data, trial_numbers


# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.

import sys
import traceback
import threading
import io
import http.server
import pickle
import json

from typing import Callable, Optional, Any

import msgpack
import numpy


class EndpointClientError(Exception):
    pass

class _NumpyJSONEncoder(json.JSONEncoder):
    """Helper class for encoding Numpy types to JSON"""
    def default(self, obj):
        if isinstance(obj, (numpy.int_, numpy.intc, numpy.intp, numpy.int8,
                            numpy.int16, numpy.int32, numpy.int64, numpy.uint8,
                            numpy.uint16, numpy.uint32, numpy.uint64)):
            return int(obj)
        elif isinstance(obj, (numpy.float_, numpy.float16, numpy.float32,
                              numpy.float64)):
            return float(obj)
        elif isinstance(obj, (numpy.ndarray,)):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)

def _numpy_msgpack_encoder(obj):
    """Helper function for encoding Numpy types to msgpack"""
    if isinstance(obj, (numpy.int_, numpy.intc, numpy.intp, numpy.int8,
                        numpy.int16, numpy.int32, numpy.int64, numpy.uint8,
                        numpy.uint16, numpy.uint32, numpy.uint64)):
        return int(obj)
    elif isinstance(obj, (numpy.float_, numpy.float16, numpy.float32,
                          numpy.float64)):
        return float(obj)
    elif isinstance(obj, (numpy.ndarray,)):
        return obj.tolist()
    return obj


def make_handler(get_callback: Optional[Callable[[io.BufferedIOBase], Any]], post_callback: Optional[Callable[[Any], Any]]):
    class Handler(http.server.BaseHTTPRequestHandler):

        def do_GET(self):
            if get_callback is None:
                self.send_error(
                    400,
                    message="This endpoint does not support GET requests")
                self.end_headers()
                return

            try:
                response = get_callback(self.rfile)
            except EndpointClientError as client_error:
                self.send_error(
                    400,
                    message=client_error.message)
                self.end_headers()
                return

            encoding_type = self.headers.get("Accept")
            if encoding_type is None or "pickle" in encoding_type:
                serialized_response = pickle.dumps(response)
            elif "json" in encoding_type:
                serialized_response = json.dumps(response, cls=_NumpyJSONEncoder).encode('UTF-8')
            elif "msgpack" in encoding_type:
                serialized_response = msgpack.packb(response, default=_numpy_msgpack_encoder)
            else:
                self.send_error(
                    400,
                    message="Unknown content type. Should be one of 'json', 'msgpack', or 'pickle'")
                self.end_headers()
                return

            self.send_response(200)
            self.end_headers()
            self.wfile.write(serialized_response)

        def do_POST(self):
            if post_callback is None:
                self.send_error(
                    400,
                    message="This endpoint does not support POST requests")
                self.end_headers()
                return

            content_length = int(self.headers.get("Content-Length", "-1"))
            serialized_body = self.rfile.read(content_length)

            encoding_type = self.headers.get("Accept")
            if encoding_type is None or "pickle" in encoding_type:
                body = pickle.loads(serialized_body)
            elif "json" in encoding_type:
                body = json.loads(serialized_body)
            elif "msgpack" in encoding_type:
                body = msgpack.unpackb(serialized_body)
            else:
                self.send_error(
                    400,
                    message="Unknown content type. Should be one of 'json', 'msgpack', or 'pickle'")
                self.end_headers()
                return

            try:
                response = post_callback(body)
            except EndpointClientError as client_error:
                self.send_error(
                    400,
                    message=str(client_error))
                self.end_headers()
                return

            # We already check that encoding_type is one of these values
            if encoding_type is None or "pickle" in encoding_type:
                serialized_response = pickle.dumps(response)
            elif "json" in encoding_type:
                serialized_response = json.dumps(response).encode('UTF-8')
            elif "msgpack" in encoding_type:
                serialized_response = msgpack.packb(response)

            self.send_response(200)
            self.end_headers()
            self.wfile.write(serialized_response)

    return Handler


class EndpointBase(threading.Thread):

    def __init__(self, endpoint_name: str, port: int, get_callback: Optional[Callable[[io.BufferedIOBase], Any]] = None, post_callback: Optional[Callable[[Any], Any]] = None):
        '''
        This class shouldn't be used directly. Use one of the implementations instead.

        Parameters
        ----------
        get_callback : Function
                       This will be called when a GET request received by the endpoint.
                       It should accept one parameter, which is a io.BufferedIOBase
                       containing the request's body.
        post_callback : Function
                        This will be called when a POST request received by the endpoint.
                        It should accept one parameter, which is the deserialized version
                        of the request's body.

        Notes
        -----
        If callbacks raise EndpointClientError, it will be send back to the client as a
        BadRequest status.
        '''
        super().__init__(daemon=True, name=endpoint_name)
        self._server = None
        self._port = port
        self._get_callback = get_callback
        self._post_callback = post_callback

    def run(self):
        try:
            self._server = http.server.ThreadingHTTPServer(
                ('0.0.0.0', self._port), make_handler(self._get_callback, self._post_callback))
            self._server.serve_forever()
        except Exception as ex:
            print("Error in {}".format(self.name), file=sys.stderr)
            traceback.print_exc()

    def stop(self):
        if self._server is not None:
            self._server.shutdown()
            self._server = None



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Foobar.
# If not, see <https://www.gnu.org/licenses/>.

import time
import http.client
import pickle
import json

import pytest
import msgpack

from octopus_sensing.device_coordinator import DeviceCoordinator
from octopus_sensing.devices.device import Device
from octopus_sensing.device_message_endpoint import DeviceMessageHTTPEndpoint
from octopus_sensing.common.message import Message


def fake_run(self):
    pass


@pytest.fixture()
def fixture():
    test_device = Device()
    test_device._run = fake_run
    coordinator = DeviceCoordinator()
    coordinator.add_device(test_device)

    endpoint = DeviceMessageHTTPEndpoint(coordinator)
    endpoint.start()

    time.sleep(0.5)

    http_client = http.client.HTTPConnection("127.0.0.1:9331", timeout=2)

    yield (endpoint, coordinator, test_device, http_client)

    endpoint.stop()


def test_should_accept_a_message(fixture):
    endpoint, coordinator, test_device, http_client, = fixture

    http_client.request(
        "POST", "/",
        body=msgpack.packb({'type': 'msg_type', 'experiment_id': '123'}),
        headers={'Accept': 'application/msgpack'})
    response = http_client.getresponse()
    assert response.status == 200

    message = test_device.message_queue.get(timeout=2)
    assert message.type == 'msg_type'
    assert message.experiment_id == '123'
    assert message.stimulus_id is None
    assert message.payload is None


def test_type_is_mandatory(fixture):
    endpoint, coordinator, test_device, http_client, = fixture

    http_client.request(
        "POST", "/",
        body=msgpack.packb({'experiment_id': '123'}),
        headers={'Accept': 'application/msgpack'})
    response = http_client.getresponse()
    assert response.status == 400


def test_json(fixture):
    endpoint, coordinator, test_device, http_client, = fixture

    http_client.request(
        "POST", "/",
        body=json.dumps({'type': 'msg_type', 'experiment_id': '123',
                         'stimulus_id': 's8', 'payload': {'data': 100}}),
        headers={'Accept': 'application/json'})
    response = http_client.getresponse()
    assert response.status == 200

    message = test_device.message_queue.get(timeout=2)
    assert message.type == 'msg_type'
    assert message.experiment_id == '123'
    assert message.stimulus_id == 's8'
    assert message.payload == {'data': 100}


def test_pickle(fixture):
    endpoint, coordinator, test_device, http_client, = fixture

    http_client.request(
        "POST", "/",
        body=pickle.dumps({'type': 'msg_type', 'experiment_id': '123',
                           'stimulus_id': 's8', 'payload': {'data': 100}}),
        headers={'Accept': 'application/pickle'})
    response = http_client.getresponse()
    assert response.status == 200

    message = test_device.message_queue.get(timeout=2)
    assert message.type == 'msg_type'
    assert message.experiment_id == '123'
    assert message.stimulus_id == 's8'
    assert message.payload == {'data': 100}


def test_bad_content_type(fixture):
    endpoint, coordinator, test_device, http_client, = fixture

    http_client.request(
        "POST", "/",
        body=pickle.dumps({'type': 'msg_type', 'experiment_id': '123',
                           'stimulus_id': 's8', 'payload': {'data': 100}}),
        headers={'Accept': 'text'})
    response = http_client.getresponse()
    assert response.status == 400


def test_multiple_endpoints(fixture):
    endpoint, coordinator, test_device, http_client, = fixture

    endpoint2 = DeviceMessageHTTPEndpoint(coordinator, port=9334)
    endpoint2.start()

    http_client.request(
        "POST", "/",
        body=msgpack.packb({'type': 'msg_type', 'experiment_id': '123'}),
        headers={'Accept': 'application/msgpack'})
    response = http_client.getresponse()
    assert response.status == 200

    message = test_device.message_queue.get(timeout=2)
    assert message.type == 'msg_type'
    assert message.experiment_id == '123'
    assert message.stimulus_id is None
    assert message.payload is None

    http_client2 = http.client.HTTPConnection("127.0.0.1:9334", timeout=2)

    http_client2.request(
        "POST", "/",
        body=msgpack.packb({'type': 'msg_type2', 'experiment_id': 'ex2'}),
        headers={'Accept': 'application/msgpack'})
    response2 = http_client2.getresponse()
    assert response2.status == 200

    message = test_device.message_queue.get(timeout=2)
    assert message.type == 'msg_type2'
    assert message.experiment_id == 'ex2'
    assert message.stimulus_id is None
    assert message.payload is None



# This file is part of Octopus Sensing <https://octopus-sensing.nastaran-saffar.me/>
# Copyright © Nastaran Saffaryazdi 2020
#
# Octopus Sensing is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
#  either version 3 of the License, or (at your option) any later version.
#
# Octopus Sensing is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Octopus Sensing.
# If not, see <https://www.gnu.org/licenses/>.
import gi
gi.require_version('Gtk', '3.0')  # nopep8
from gi.repository import Gtk  # nopep8
from typing import List, Union

from octopus_sensing.questionnaire.question import Question

FONT_STYLE = "<span font_desc='Tahoma 16'>{}</span>"


class TextQuestion(Question):
    '''
    The class for creating text questions using Gtk 3.0

    Attributes
    ----------

    Parameters
    ----------
    id: str
        A unique ID for the question
    
    text: str
        The text of question
    
    default_answer: Union[int, str], default: 0
        The default answer
    
    '''
    def __init__(self, id: str, text: str, default_answer: Union[int, str] = 0):
        super().__init__(id, text)

        self.answer_textbox = Gtk.Entry()
        self.answer_textbox.set_text(default_answer)

    def render(self, grid: Gtk.Grid, grid_row: int) -> int:
        '''
        renders a question for adding to a questionnaire

        Parameters
        ----------
        grid: Gtk.Grid
            a Gtk grid object that this question will be added to it
            
        grid_row: int
            The row number of grid that the question will be added to it
        
        Returns
        -------
        row_counter: int
            The grid's row for adding the next object after adding this question
        
        Examples
        --------
        Creating a text question and adding it to the questionnaire

        >>> question_1 = TextQuestion("q1",
        ...                           "1- What emotion did you feel the most?",
        ...                           default_answer="Happiness")
        >>> questionnaire = Questionnaire("after_stimuli",
        ...                               "study01_p10",
        ...                               "stimuli00",
        ...                               "After Stimulus Questionnaire")
        >>> questionnaire.add_question(question_1)

        See Also
        -----------
        :class:`octopus_sensing.questionnaire.questionnaire`
        
        '''
        row_counter = grid_row

        # Question box
        question_label_box = Gtk.Box(spacing=120)
        question_label = Gtk.Label()
        question_label.set_markup(FONT_STYLE.format(self._text))
        question_label_box.pack_start(question_label, False, False, 0)

        grid.attach(question_label_box, 0, row_counter, 1, 1)
        row_counter += 1

        # text box
        text_box = Gtk.Box(spacing=120)
        text_box.pack_start(self.answer_textbox, False, False, 0)
        grid.attach(text_box, 1, row_counter, 1, 1)
        row_counter += 1
        return row_counter

    def get_answer(self) -> Union[int, str]:
        '''
        Gets the answer

        Returns
        ----------
        answer: str or int
            answer
        '''
        return self.answer_textbox.get_text()
